/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 2001-2002 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */

#include "nv-misc.h"
#include "os-interface.h"
#include "nv.h"
#include "nv-freebsd.h"

static d_open_t  nvidia_ctl_open;
static d_close_t nvidia_ctl_close;
static d_ioctl_t nvidia_ctl_ioctl;
static d_poll_t  nvidia_ctl_poll;

static struct cdevsw nvidia_ctl_cdevsw = {
#if __FreeBSD_version < 500105
    /* open    */  nvidia_ctl_open,
    /* close */    nvidia_ctl_close,
    /* read */     noread,
    /* write */    nowrite,
    /* ioctl */    nvidia_ctl_ioctl,
    /* poll */     nvidia_ctl_poll,
    /* mmap */     nommap,
    /* strategy */ nostrategy,
    /* name */     "nvidiactl",
    /* maj */      CDEV_MAJOR,
    /* dump */     nodump,
    /* psize */    nopsize,
    /* flags */    D_TRACKCLOSE,
#if __FreeBSD_version < 500000
    /* bmaj */     -1,
#endif
    /* kqfilter */ NULL,
#else
    .d_open =      nvidia_ctl_open,
    .d_close =     nvidia_ctl_close,
    .d_ioctl =     nvidia_ctl_ioctl,
    .d_poll =      nvidia_ctl_poll,
    .d_name =      "nvidiactl",
#if __FreeBSD_version >= 502103
    .d_version =   D_VERSION,
    .d_flags =     D_TRACKCLOSE|D_NEEDGIANT
#else
    .d_maj =       CDEV_MAJOR,
    .d_flags =     D_TRACKCLOSE
#endif /* XXX Fix me? (fbsd version) */
#endif
};

static dev_t          nvidia_ctl_cdev;
static nvidia_softc_t nvidia_ctl_sc;
static int            nvidia_count = 0;

int nvidia_ctl_open(
    dev_t dev,
    int oflags,
    int devtype,
    d_thread_t *td
)
{
    int status;
    nv_state_t *nv = &nvidia_ctl_state;

    nv_lock_api(nv);
    status = nvidia_open_ctl(dev, td);
    nv_unlock_api(nv);

    return status;
}

int nvidia_ctl_close(
    dev_t dev,
    int fflag,
    int devtype,
    d_thread_t *td
)
{
    int status;
    nv_state_t *nv = &nvidia_ctl_state;

    nv_lock_api(nv);
    status = nvidia_close_ctl(dev, td);
    nv_unlock_api(nv);

    return status;
}

int nvidia_ctl_ioctl(
    dev_t dev,
    u_long cmd,
    caddr_t data,
    int fflag,
    d_thread_t *td
)
{
    int status = 0;
    nv_state_t *nv = &nvidia_ctl_state;

    if (__NV_IOC_TYPE(cmd) != NV_IOCTL_MAGIC)
        return ENOTTY;

    nv_lock_api(nv);

    switch (cmd) {
        case NV_IOCTL_CARD_INFO:
            do {
                struct nv_ioctl_card_info *ci = (void *) data;
                status = nvidia_get_card_info(ci);
            } while (0);
            break;

        case NV_IOCTL_POST_VBIOS:
            do {
                struct nv_ioctl_post_vbios *pc = (void *) data;
                status = nvidia_post_vbios(pc);
            } while (0);
            break;

        case NV_IOCTL_ENV_INFO:
            do {
                struct nv_ioctl_env_info *ei = (void *) data;
                ei->sim_env = nv->sim_env;
                ei->pat_supported = 0; /* for now */
            } while (0);
            break;

        case NV_IOCTL_CHECK_VERSION_STR:
            do {
                nv_ioctl_rm_api_version_t *params = (void *) data;
                status = (rm_perform_version_check(params) == RM_OK) ? 0 : -EINVAL;
            } while (0);
            break;

        default:
            status = nvidia_handle_ioctl(dev, cmd, data, fflag, td);
    }

    nv_unlock_api(nv);
    return status;
}


/*
 * The X driver module adds the ctrl device fd to the XFree86 select list
 * and receives notification if the select status changes. We make use of
 * this mechanism to signal hotkey events; this function is called by the
 * core resource manager whenever a legacy hotkey event is detected.
 */

void nv_set_hotkey_occurred_flag(void)
{
    nv_state_t *nv;
    struct nvidia_softc *sc;

    nv = &nvidia_ctl_state;
    sc = nv->os_state;

    nv_lock_rm(nv);
    nv->flags |= NV_FLAG_HOTKEY_OCCURRED;
    nv_unlock_rm(nv);

    selwakeup(&sc->rsel);
}

int nvidia_ctl_poll(
    dev_t dev,
    int events,
    d_thread_t *td
)
{
    nv_state_t *nv;
    struct nvidia_softc *sc;

    nv = &nvidia_ctl_state;
    sc = nv->os_state;

    nv_lock_rm(nv);

    if ((nv->flags & NV_FLAG_HOTKEY_OCCURRED) == 0) {
        nv_unlock_rm(nv);
        selrecord(td, &sc->rsel);
    } else {
        nv->flags &= ~NV_FLAG_HOTKEY_OCCURRED;
        nv_unlock_rm(nv);

        return (events & (POLLIN | POLLRDNORM)); /* ? */
    }

    return 0;
}

int nvidia_ctl_attach(void)
{
    struct nvidia_softc *sc;

    if (nvidia_count == 0) {
        /*
         * This routine is called from nvidia_attach, multiple times when
         * more than one device is installed - this makes it necessary to
         * prevent multiple calls to make_dev.
         */
        nvidia_ctl_cdev = make_dev(&nvidia_ctl_cdevsw,
                CDEV_CTL_MINOR,
                UID_ROOT, GID_WHEEL, 0666,
                "%s", nvidia_ctl_cdevsw.d_name);

        sc = &nvidia_ctl_sc;
        bzero(sc, sizeof(nvidia_softc_t));
        nvidia_ctl_state.os_state = sc;

        STAILQ_INIT(&sc->filep_queue);

#if __FreeBSD_version >= 500000
        mtx_init(&sc->mtx_rm, "ctl.mtx_rm", NULL, MTX_SPIN | MTX_RECURSE);
        sx_init(&sc->sx_api, "ctl.sx_api");
#else
        /*
         * Assume that none of the API paths are reentrant and
         * that any could could conceivably sleep.
         */
        lockinit(&sc->api_lock, PZERO | PCATCH, "ctl.api_lock", 0, 0); 
#endif
    }

    nvidia_count++;
    return 0;
}

int nvidia_ctl_detach(void)
{
    struct nvidia_softc *sc;

    sc = &nvidia_ctl_sc;
    nvidia_count--;

    if (nvidia_count == 0) {
        /*
         * Similar to nvidia_ctl_attach, nvidia_ctl_detach has to respect
         * the device reference count to prevent premature destruction of
         * the control device.
         */
#if __FreeBSD_version >= 500000
        mtx_destroy(&sc->mtx_rm);
        sx_destroy(&sc->sx_api);
#else
        lockmgr(&sc->api_lock, LK_DRAIN, 0, CURTHREAD);
#endif
        destroy_dev(nvidia_ctl_cdev);
    }

    return 0;
}

